#include <u.h>
#include <libc.h>
#include <bio.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include "../lib/snmp.h"
#include "../lib/snmp_debug.h"
/* Optional: #include "beer.h" or #include "caffeine.h" */

/* creates a new snmp request id */
/* based on rsc code */
int
mkreqid(void) 
{
	snmpdbg(DEBUG, GOBACK, "mkreqid()", "Creating a Request ID.");
	return (nsec() & 0x7FFE) + 1;
}


Tuple *
mktuple(char * oid, char * value, char * type, int rw)
{
	Tuple *aux = nil;
	char *moid = nil, *mvalue = nil, *mtype = nil;

	snmpdbg(DEBUG, PLUS, "mktuple()", "Creating a Tuple.");

	if (oid == nil) {
		snmpdbg(DEBUG, MINUS, "mktuple()", "No OID set.");
		return(nil);
	}
	moid = strdup(oid);

	if (rw == 1) {
		/* SNMP Set Request */
		if ((value == nil) || (type == nil)) {
			snmpdbg(DEBUG, MINUS, "mktuple()", "No Value o Type set.");
			return(nil);
		}
		mvalue = strdup(value);
		mtype = strdup(type);
	} else {
		/* SNMP  Get's Request */
		if (value == nil) {
			mvalue = nil;
		} else {
			mvalue = strdup(value);
		}

		if (type == nil) {
			type = nil;
		} else {
			mtype =  strdup(type);
		}
	}

	aux = (Tuple*) malloc(sizeof(Tuple));
	aux->oid = moid;
	aux->value = mvalue;
	aux->type = mtype;
	aux->nexttuple = nil;

	snmpdbg(DEBUG, MINUS, "mktuple()", "Created. addr = %p", aux);

	return aux;
}


int
linktuples(Tuple *father, Tuple *child)
{
	Tuple *auxtpl;

	snmpdbg(DEBUG, PLUS, "linktuples()", "Linking tuples.");

	auxtpl = father;

	while(auxtpl->nexttuple != nil) {
		auxtpl = auxtpl->nexttuple;
	}

	auxtpl->nexttuple = child;

	snmpdbg(DEBUG, MINUS, "linktuples()", "Linked.");

	return 0;
}


void
freetuple(Tuple *t)
{
	if (t != nil) { 
		snmpdbg(DEBUG, PLUS, "freetuple()", "Freeing the memory of a Tuple.");

		if (t->nexttuple != nil) {
			/* go to the next node */
			freetuple(t->nexttuple);
		}

		/* Now, free the node */
		if (t->oid != nil) {
			snmpdbg(DEBUG, EQUAL, "freetuple()", "OID=%s", t->oid);
			free(t->oid);
		}

		if (t->value != nil) {
			snmpdbg(DEBUG, EQUAL, "freetuple()", "Val=%s", t->value);
			free(t->value);
		}

		if (t->type != nil) {
			snmpdbg(DEBUG, EQUAL, "freetuple()", "Type=%s", t->type);
			free(t->type);
		}

		snmpdbg(DEBUG, EQUAL, "freetuple()", "Freeing the Tuple. addr = %p", t);
		free(t);

		snmpdbg(DEBUG, MINUS, "freetuple()", "Freeded.");
	} else {
		snmpdbg(DEBUG, EQUAL, "freetuple()", "The tuple is already nil");
	}
}


Pdu*
mkpdu(void)
{
	Pdu *pdu = nil;

	snmpdbg(DEBUG, PLUS, "mkpdu()", "Creating a Pdu.");
	pdu = (Pdu*) malloc(sizeof(Pdu));

	if (pdu == nil) {
		snmpdbg(DEBUG, MINUS, "mkpdu()", "No memory.");
		return(nil);
	}

	pdu->tuple = nil;

	snmpdbg(DEBUG, EQUAL, "mkpdu()", "Memory for the PDU allocated. addr = %p", pdu);
	snmpdbg(DEBUG, MINUS, "mkpdu()", "Created.");

	return(pdu);
}


void
freepdu(Pdu *p)
{
	if (p != nil) {
		snmpdbg(DEBUG, PLUS, "freepdu()", "Freeing the memory of a PDU.");

		freetuple(p->tuple);
		snmpdbg(DEBUG, EQUAL, "freepdu()", "Freeing the PDU. addr = %p", p);
		free(p);

		snmpdbg(DEBUG, MINUS, "freepdu()", "Freeded.");
	} else {
		snmpdbg(DEBUG, EQUAL, "freepdu()", "The PDU is already nil.");
	}
}


Packet*
mkpkt(void) 
{
	Packet  *pkt = nil;

	snmpdbg(DEBUG, PLUS, "mkpkt()", "Creating a Packet.");

	/* Get memory to create the packet */
	pkt = (Packet*) malloc(sizeof(Packet));
	if (pkt == nil) {
		snmpdbg(DEBUG, EQUAL, "mkpkt()", "No memory for the packet");
	} else {
		pkt->comm = nil;
		pkt->pdu = nil;
		pkt->nextpkt = nil;

		snmpdbg(DEBUG, EQUAL, "mkpkt()", "Memory for the packet allocated. addr = %p", pkt);
		snmpdbg(DEBUG, MINUS, "mkpkt()", "Created.");
	}

	return(pkt);
}


int
linkpkt(Packet *father, Packet *child)
{
	Packet *auxpkt;

	snmpdbg(DEBUG, PLUS, "linkpkt()", "Linking packets.");

	auxpkt = father;

	while(auxpkt->nextpkt != nil) {
		auxpkt = auxpkt->nextpkt;
	}

	auxpkt->nextpkt = child;

	snmpdbg(DEBUG, MINUS, "linkpkt()", "Linked.");

	return(0);
}


int
fillpkt(Packet * pkt, snmp_session *sess, int type, char *oids, char *values, char *types, int nonrep, int maxrep)
{
	char **loids = nil, **lvals = nil, **ltyps = nil;
	char *tmpoids = nil, *tmpvals = nil, *tmptyps = nil;
	Pdu *pdu;
	Tuple *tmptpl;
	int noid = 0, nval = 0, ntyp = 0, i = 0;

	if (sess != nil) {
		snmpdbg(DEBUG, PLUS, "fillpkt()", "The Packet contains data.");

		if ((type == GetBulkRequest) && (sess->version < SNMP_VERSION_2)) {
			snmpdbg(DEBUG, MINUS, "fillpkt()", "GetBulk in SNMPv1 version.");
			return(-1);
		} 

		if ((oids == nil) || ((type == SetRequest) && ((values == nil) || (types == nil)))) {
			snmpdbg(DEBUG, MINUS, "fillpkt()", "The OID List is nil, or Values (in Set) is nil");
			return(-1);
		}
	}

	snmpdbg(DEBUG, EQUAL, "fillpkt()", "Preparing the Request.");

	/* Create the PDU */
	if (pkt == nil) {
		snmpdbg(DEBUG, MINUS, "fillpkt()", "The packet is nil.");
		return(-2);
	}

	pdu = mkpdu();

	if (pdu == nil) {
		snmpdbg(DEBUG, MINUS, "fillpkt()", "Not PDU created.");
		return(-2);
	}

	pkt->pdu = pdu;

	/* Set the values in the packet */
	pkt->pdu->type = type;
	pkt->pdu->reqid = mkreqid();
	pkt->pdu->errstat = 0;
	pkt->pdu->erridx = 0;
	pkt->pdu->tlen = noid;
	pkt->pdu->nrep = nonrep;
	pkt->pdu->maxrep = maxrep;

/************************************/

	/* Preparing the Object ID List (loids) to set up with the array from "oids" */
	loids = (char**) malloc(sizeof(char*) * sess->maxoids);

	/* We must save the values of the oids, not modify it in the getfields */
	tmpoids = strdup(oids);

	if ((loids == nil) || (tmpoids == nil)) {
		snmpdbg(DEBUG, MINUS, "fillpkt()", "The OID List is nil");
		return(-2);
	}

	noid = getfields(tmpoids, loids, sess->maxoids, 0, ",");

	if (type == SetRequest) {
		/* We must save the values to not modify it in the getfields */
		snmpdbg(DEBUG, EQUAL, "fillpkt()", "Preparing the SetRequest.");
		lvals = (char**) malloc(sizeof(char*) * sess->maxoids);
		tmpvals = strdup(values);

		ltyps = (char**) malloc(sizeof(char*) * sess->maxoids);
		tmptyps = strdup(types);

		if ((lvals == nil) || (tmpvals == nil)) {
			snmpdbg(DEBUG, MINUS, "fillpkt()", "The Value List is nil");
			return(-2);
		}

		if ((ltyps == nil) || (tmptyps == nil)) {
			snmpdbg(DEBUG, MINUS, "fillpkt()", "The Value Type's List is nil");
			return(-2);
		}

		nval = getfields(tmpvals, lvals, sess->maxoids, 0, ",");
		ntyp = getfields(tmptyps, ltyps, sess->maxoids, 0, ",");

		if ((nval != noid) || (nval != ntyp)) {
			snmpdbg(DEBUG, EQUAL, "fillpkt()", "In a SetRequest the OIDs, Values and Types must be equals.");
			free(loids);
			free(tmpoids);
			free(lvals);
			free(tmpvals);
			free(ltyps);
			free(tmptyps);

			return(-1);
		}
	}

	/* Prepare the oids */
	/* Must be at least one oid for the get or two for the set */
	if (pkt->pdu->type == SetRequest) {
		pkt->pdu->tuple = mktuple(loids[0], lvals[0], ltyps[0], 1);
	} else {
		pkt->pdu->tuple = mktuple(loids[0], "nil", "nil", 0);
	}

	if (pkt->pdu->tuple == nil) {
		snmpdbg(DEBUG, MINUS, "fillpkt()", "The tuple is nil.");
		free(loids);
		free(tmpoids);
		if (type == SetRequest) {
			free(lvals);
			free(tmpvals);
			free(ltyps);
			free(tmptyps);
		}
		return(-2);
	}

	/* The next oids */
	for(i = 1; i < noid; i++) {
		if (pkt->pdu->type == SetRequest) {
			tmptpl = mktuple(loids[i], lvals[i], ltyps[i], 1);
		} else {
			tmptpl = mktuple(loids[i], "nil", "nil", 0);
		}

		if (tmptpl == nil) {
			snmpdbg(DEBUG, MINUS, "fillpkt()", "The tuple is nil.");
			free(loids);
			free(tmpoids);
			if (type == SetRequest) {
				free(lvals);
				free(tmpvals);
				free(ltyps);
				free(tmptyps);
			}
			return(-2);
		}
		linktuples(pkt->pdu->tuple, tmptpl);
	}

	snmpdbg(DEBUG, MINUS, "fillpkt()", "All Ok, cleaning memory.");

	free(loids);
	free(tmpoids);
	if (type == SetRequest) {
		free(lvals);
		free(tmpvals);
		free(ltyps);
		free(tmptyps);
	}

	return(0);
}


void
freepkt(Packet *p)
{
	snmpdbg(DEBUG, PLUS, "freepkt()", "Freeing the memory of a Packet.");

	if (p != nil) {

		if (p->nextpkt != nil) {
			/* go to the next node */
			freepkt(p->nextpkt);
		}

		/* Now, with the packet, first the PDU */
		freepdu(p->pdu);
		if (p->comm != nil) {
			snmpdbg(DEBUG, EQUAL, "freepkt()", "Freeing the Packet-> community.");
			free(p->comm);
		}

		snmpdbg(DEBUG, EQUAL, "freepkt()", "Freeing the Packet. addr = %p", p);
		free(p);
	} else {
		snmpdbg(DEBUG, EQUAL, "freepkt()", "The packet is nil, nothing to do.");
	}

	snmpdbg(DEBUG, MINUS, "freepkt()", "Freeded.");
}

Packet *
snmpDoIt(snmp_session *s, Packet *snd, Packet *rcv)
{
	snmpdbg(DEBUG, PLUS, "snmpDoIt()", "Doing a SNMP Get.");

	if ( snmpTalk(s, snd, rcv) < 0 ) {
		snmpdbg(DEBUG, EQUAL, "snmpDoIt()", "error talking snmp.");
		rcv = nil;
	}

	snmpdbg(DEBUG, MINUS, "snmpDoIt()", "SNMP Get Done.");
	return(rcv);
}


char * 
dumppkt(Packet *pkt)
{
	char *line, *buff, *ptr;
	int n, linelen, len = 0;
	Tuple *tmptpl;
	Packet *tmppkt;

	snmpdbg(DEBUG, PLUS, "dumppkt()", "Dumping the packets.");

	if (pkt == nil) {
		snmpdbg(DEBUG, MINUS, "dumppkt()", "The packet is nil.");
		return(nil);
	}

	buff = malloc(0);
	tmppkt = pkt;

	while (tmppkt != nil) {
		snmpdbg(DEBUG, EQUAL, "dumppkt()", "Dumping a packet.");
		tmptpl = tmppkt->pdu->tuple;

		while (tmptpl != nil) {
			snmpdbg(DEBUG, EQUAL, "dumppkt()", "Dumping a tuple.");
			linelen = strlen(tmptpl->oid) + strlen(tmptpl->value) + strlen(tmptpl->type) + 11;
			buff = erealloc9p(buff, len + linelen);
			line = malloc(linelen);
			n = snprint(line, linelen + 11, "[(%s) (%s) (%s)]", tmptpl->oid, tmptpl->value, tmptpl->type);
			snmpdbg(DEBUG, EQUAL, "dumppkt()", "[(%s) (%s)]", tmptpl->oid, tmptpl->value);
			ptr = buff + len;
			memcpy(ptr, line, n);
			len += n;

			free(line);

			tmptpl = tmptpl->nexttuple;
		}
		tmppkt = tmppkt->nextpkt;
	}

	buff[len] = '\0';

	snmpdbg(DEBUG, MINUS, "dumppkt()", "Packets dumped.");
	
	return buff;
}


/*
  * EXTERNAL FUNCTIONS, RETURN CHAR POINTERS 
  ********* */
char *
snmpCmd(snmp_session *sess, char *oids, char *values, 
		char *types, int norep, int maxrep, int snmpType)
{
	char *buff = nil;
	Packet *req = nil, *rcv = nil;
	int rtr = 0;

	req = mkpkt();
	if (req == nil) {
		return(nil);
	}

	rcv = mkpkt();
	if (req == nil) {
		freepkt(req);
		return(nil);
	}

	switch(snmpType) {
		case GetRequest:
		case GetNextRequest:
		case SetRequest:
		case GetBulkRequest:
			rtr = fillpkt(req, sess, snmpType, oids, values, types, norep, maxrep);
			if (rtr != 0) {
				freepkt(req);
				freepkt(rcv);
				return(nil);
			}

			buff = dumppkt(snmpDoIt(sess, req, rcv));
			break;
		case SnmpWalk:
			fillpkt(req, sess, GetRequest, oids, values, types, norep, maxrep);
			if (rtr != 0) {
				freepkt(req);
				freepkt(rcv);
				return(nil);
			}

			buff = dumppkt(walkit(sess, req, rcv));
			break;		
	}

	freepkt(req);
	freepkt(rcv);

	return buff;
}

/* ************************
  *  SNMP Walk Functions
  * ************************/
Packet *
walkit(snmp_session *sess, Packet *pktreq, Packet *pktrcv)
{
	Packet *tosnd, *torcv, *header, *prelast;
	int errflg = 0;
	char *oidchar;
	int round = 0;
	int rtr = 0;
	int goon = 0;

	snmpdbg(DEBUG, PLUS, "walkit()", "Doing a SNMP Walk.");

	/* Preparing the header of the packet list */
	header = nil;
	prelast = nil;

	/* Preparing the first packet to send */
	tosnd = pktreq;
	torcv = pktrcv;

	do {
		if ( snmpDoIt(sess, tosnd, torcv) == nil ) {
			snmpdbg(DEBUG, EQUAL, "walkit()", "Error talking snmp.");
			errflg = 1;
		} else {
			/* Save the packet in a list 
			  * except if is the first response
			  * beacuse is the same packet
			  */
			if (round == 0) {
				if (isnotsuchobj(torcv) != 0) {
					/* Probably you select a branch, not a leaf */
					/* Try with the GetNextRequest to select a leaf */
					/* Then, we do not save the leaf in the tree, and repeat with the next one */
					snmpdbg(DEBUG, EQUAL, "walkit()", "Branch, not leaf");
					round = 0;
					goon = 1;

					/* Get the next oid to ask */
					oidchar = torcv->pdu->tuple->oid;

					/* Clean the packet tosnd contents */
					freepdu(tosnd->pdu);
					if (tosnd->comm != nil) {
						free(tosnd->comm);
					}

					rtr = fillpkt(tosnd, sess, GetNextRequest, oidchar, nil, nil, 0, 0);
					if (rtr != 0) {
						errflg = 1;
					}

					/* Clean the packet torcv contents */
					freepdu(torcv->pdu);
					if (torcv->comm != nil) {
						free(torcv->comm);
					}

				} else {
					snmpdbg(DEBUG, EQUAL, "walkit()", "Leaf");
					if (iseom(torcv) != 0) {
						/* Yeah!, you select the last leaf in the tree */
						/* Then, save it and not continue */
						goon = 0;
						header = torcv;
						prelast = torcv;
					} else {
						/* This is the normal way, a leaf and not the last one */
						/* Then, save it and continue */
						goon = 1;
						header = torcv;
						prelast = torcv;
						/* OK, exists!!, we can save it, then we do not enter in the first round anymore */
						round = 1;


						/* Get the next oid to ask */
						oidchar = torcv->pdu->tuple->oid;
		
						/* Fill the new packet to send */

						/* First, clean the contents */
						freepdu(tosnd->pdu);
						if (tosnd->comm != nil) {
							free(tosnd->comm);
						}

						/* Now, refill it */
						rtr = fillpkt(tosnd, sess, GetNextRequest, oidchar, nil, nil, 0, 0);
						if (rtr != 0) {
							errflg = 1;
						}

						/* Create the new packet to recive */
						torcv = mkpkt();

					}
				}
			} else {
				/* Link the packet */
				linkpkt(header, torcv);

				goon = (iseom(torcv) == 0); // ( || (isnotsuchobj(pktrcv)) !=0);

				/* Prepare the next comparation */
				prelast = torcv;

				/* Get the next oid to ask */
				oidchar = torcv->pdu->tuple->oid;
		
				/* Fill the new packet to send */

				/* First, clean the contents */
				freepdu(tosnd->pdu);
				if (tosnd->comm != nil) {
					free(tosnd->comm);
				}

				/* Now, refill it */
				rtr = fillpkt(tosnd, sess, GetNextRequest, oidchar, nil, nil, 0, 0);
				if (rtr != 0) {
					errflg = 1;
				}

				/* Create the new packet to recive */
				if (goon != 0) {
					torcv = mkpkt();
				}
			}
		}
//				if ((canwalk(prelast, torcv) != 0) || (iseom(torcv) != 0)) {

	} while( goon && !errflg );
	snmpdbg(DEBUG, MINUS, "walkit()", "SNMP Walk done.");
	return header;
}


/*
  * Values:
  * 0 = OK, are different
  * 1 = OK, are equal, not continue
  * 2 = Problem
  */
int
canwalk(Packet *prelast, Packet *torcv)
{
	if ((prelast == nil) || (torcv == nil)) {
		return(2);
	}

	if (strcmp(prelast->pdu->tuple->oid, torcv->pdu->tuple->oid) ) {
		return(0);
	} else {
		return(1);
	}
}


/*
  * Values:
  * 0 = OK, not EndOfMib
  * 1 = OK, EndOfMib
  * 2 = Problem
  */
int
iseom(Packet *pkt)
{
	char mychar[] = "End of MIB";

	if(pkt != nil) {
		if (strcmp(pkt->pdu->tuple->value, mychar)) {
			return(0);
		} else {
			return(1);
		}
	} else {
		return(2);
	}
}

/*
  * Values:
  * 0 = OK, not NoSuchObject/NoSuchInstance
  * 1 = OK, NoSuchObject/NoSuchInstance
  * 2 = Problem
  */
int
isnotsuchobj(Packet *pkt)
{
	char mychar1[] = "No Such Object";
	char mychar2[] = "No Such Instance";

	if(pkt != nil) {
		if ((strcmp(pkt->pdu->tuple->value, mychar1)) && (strcmp(pkt->pdu->tuple->value, mychar2))) {
			return(0);
		} else {
			return(1);
		}
	} else {
		return(2);
	}
}