#include <u.h>
#include <libc.h>
#include <bio.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include "../lib/snmp.h"
#include "../lib/snmp_debug.h"
/* Global variables for the timeout */
int alarmflag;
int readflag;

/* 
  * Timeout Code from RSC
  * Probably, if the snmpfs fails is because this code is very bad :-)
  */
int
alarmtr(void *v, char *why)
{
	snmpdbg(DEBUG, PLUS, "alarmtr()", "Alarm captured.");

	if(!readflag) {
		snmpdbg(DEBUG, MINUS, "alarmtr()", "Alarm served (exit 0).");
		return 0;
	}

	if(strcmp(why, "alarm") == 0) {
		alarmflag++;
		snmpdbg(DEBUG, MINUS, "alarmtr()", "Alarm served (exit 1).");
		return 1;
	}

	snmpdbg(DEBUG, MINUS, "alarmtr()", "Alarm served (exit 0).");
	return 0;
}


/* ASN.1 VarBind Hard Encapsulation */
/*
  * Var-Bind-List =SEQUENCE 		
  *	Var-Bind = SEQUENCE		
  *		oid	OBJECT IDENTIFIER	
  *		oid	OBJECT IDENTIFIER
  *		NULL
  */
Elem
mkvarbind(Tuple *tuple)
{
	Elist *x, *y, *z = nil;
	Tuple *tmptpl, *storetpl, *newtpl, *revtpl;
	Elem oid, value;

	tmptpl = tuple;

	/* The mkel() function store the elements in reverse order
	  * therefore, I create a temporal list in reverse order too.
	  * One element is needed to make the header
	  */
	revtpl = mktuple(tmptpl->oid, tmptpl->value, tmptpl->type, 2);
	tmptpl = tmptpl->nexttuple;

	/* When the header is created, continue with the next elements */
	while(tmptpl != nil) {
		newtpl = mktuple(tmptpl->oid, tmptpl->value, tmptpl->type, 2);
		tmptpl = tmptpl->nexttuple;

		/* Link the tuple to the header, relink the header and go! */
		linktuples(newtpl, revtpl);
		revtpl = newtpl;
	}

	/* Here we have two lists, tuple (original), storetpl (reverse) */
	storetpl = revtpl;

	while(storetpl != nil) {
		value = char2asn(storetpl->value, storetpl->type);
		oid     = oid2asn(storetpl->oid);
		x = mkel(value, nil);
		y = mkel(oid, x);
		z = mkel(mkseq(y), z);

		storetpl = storetpl->nexttuple;
	}

	/* Now, destroy the temporal (reverse) list */
	freetuple(revtpl);

	return mkseq(z);
}


/* ASN.1 Bulk Hard encapsulation */
Elist *
snmpbulk(Packet *pkt)
{
	Elist *e = nil;

	e = mkel(mkSNMPpdu(pkt->pdu->type,
		mkel(mkint(pkt->pdu->reqid),
		mkel(mkint(pkt->pdu->nrep),
		mkel(mkint(pkt->pdu->maxrep),
		mkel(mkvarbind(pkt->pdu->tuple),nil))))),nil);

	return(e);
}


/* ASN.1 Request Hard encapsulation */
Elist *
snmprequest(Packet *pkt)
{
	Elist *e = nil;

	e = mkel(mkSNMPpdu(pkt->pdu->type,
		mkel(mkint(pkt->pdu->reqid),
		mkel(mkint(pkt->pdu->errstat),
		mkel(mkint(pkt->pdu->erridx),
		mkel(mkvarbind(pkt->pdu->tuple),nil))))),nil);

	return(e);
}

Elem
mkSNMPpdu(int type, Elist *el)
{
	Elem e;

	e.tag.class = type;
	e.tag.num = SEQUENCE;
	e.val.tag = VSeq;
	e.val.u.seqval = el;

	return e;
}


/* ASN.1 snmp packet encoding
  * 
  * message = SEQUECE
  *		version INTEGER
  *		community OCTECT_STRING
  *		PDU = IMPLICIT SEQUENCE
  */
int
encpkt(snmp_session *sess, Packet *pkt, Bytes **pbytes)
{
	Elem snmp, comm, ver;
	Elist * pdu;
	int err;
	Bytes *bsnmp;
	
	snmpdbg(DEBUG, PLUS, "encpkt()", "Encapsulating the Packet.");

	if ((pkt == nil) || (pkt->pdu == nil)) {
		snmpdbg(DEBUG, MINUS, "encpkt()", "The Packet or the PDU is a null pointer.");
		return(-2);
	}

	switch(pkt->pdu->type) {
		case GetRequest: 
			snmpdbg(DEBUG, EQUAL, "encpkt()", "Making a SNMP GetRequest.");
			pdu = snmprequest(pkt);
			break;
		case GetNextRequest:
			snmpdbg(DEBUG, EQUAL, "encpkt()", "Making a SNMP NextRequest.");
			pdu = snmprequest(pkt);
			break;
		case SetRequest:
			snmpdbg(DEBUG, EQUAL, "encpkt()", "Making a SNMP SetRequest.");
			pdu = snmprequest(pkt);
			break;
		case GetBulkRequest:
			snmpdbg(DEBUG, EQUAL, "encpkt()", "Making a SNMP BulkRequest.");
			pdu = snmpbulk(pkt);
			break;
		default:
			return(-1);
			break;
	}

	if ( pdu == nil ) {
		snmpdbg(DEBUG, MINUS, "encpkt()", "The PDU is a null pointer.");
		return(-2);
	}

	ver = mkint(sess->version);

	snmpdbg(DEBUG, EQUAL, "encpkt()", "The PDU contains data.");

	if (pkt->pdu->type == SetRequest) {
		comm = mkoctet((uchar*)sess->rwcommunity, sess->rwcommlen);
	} else {
		comm = mkoctet((uchar*)sess->rocommunity, sess->rocommlen);
	}

	snmp = mkseq(mkel(ver, mkel(comm, pdu)));


	err = encode(snmp, &bsnmp);
	if ( err != ASN_OK ) {
		snmpdbg(DEBUG, MINUS, "encpkt()", "Cannot encode the pkt: %d",err);
		return(-5);
	}

	*pbytes = bsnmp;

	/* Now, free the list */
	snmpdbg(DEBUG, EQUAL, "encpkt()", "Free the old encapsulation PDU.");
	freevalfields(&snmp.val);

	snmpdbg(DEBUG, MINUS, "encpkt()", "Packet encapsulation Success.");
	return(err);
}

int
snmpTalk(snmp_session *sess, Packet *snd, Packet *rcv)
{
	int n, retries, reqid, fd;
	uchar *rcvbuf;
	Bytes *snmp;

	snmpdbg(DEBUG, PLUS, "snmpTalk()", "Entering in the talk.");

	/* Make the encapsulation in ASN.1 */
	snmpdbg(DEBUG, EQUAL, "snmpTalk()", "Encapsulating in ASN1.");
	if (encpkt(sess, snd, &snmp)) {
		snmpdbg(DEBUG, MINUS, "snmpTalk()", "Error encoding Packet to ASN1.");
		return -4;
	}

	/* Prepare the reception buffer */
	rcvbuf = (uchar*) malloc(sess->rcvMsgMaxSize);

	if (rcvbuf == nil) {
		snmpdbg(DEBUG, MINUS, "snmpTalk()", "No memory available for the reception buffer.");
		freebytes(snmp);
		return -2;
	}

	/* Prepare the local variables, from the global */
	reqid = snd->pdu->reqid;
	retries = sess->retries;

	/* set alarm to exit the while when timeout reached */
	alarmflag = 0;
	readflag = 1;
	atnotify(alarmtr, 1);
	alarm(sess->timeout);
	snmpdbg(DEBUG, EQUAL, "snmpTalk()", "timeout = %d", sess->timeout);

	/* Prepare the little file descriptor */
	snmpdbg(DEBUG, EQUAL, "snmpTalk(): Dialing to %s", sess->host);
	if ((fd = dial(netmkaddr(sess->host, "udp", sess->port), 0, 0, 0))< 0) {
		snmpdbg(DEBUG, MINUS, "snmpTalk(): error dialing ", sess->host);
		freebytes(snmp);
		free(rcvbuf);
		return -2;
	}

	/* set alarm to exit the while when timeout reached */
	alarm(0);
	
	do {
		n = write(fd, snmp->data, snmp->len);
		
		if (n != snmp->len) {
			snmpdbg(DEBUG, MINUS, "snmpTalk()", "Error writing to %d", fd);
			freebytes(snmp);
			free(rcvbuf);
			return -3;
		}

		/* set alarm to exit the while when timeout reached */
		alarmflag = 0;
		readflag = 1;
		atnotify(alarmtr, 1);
		alarm(sess->timeout);
		snmpdbg(DEBUG, EQUAL, "snmpTalk()", "timeout = %d", sess->timeout);
		do {
			n = read(fd, rcvbuf, sess->rcvMsgMaxSize);

			if (n > 0)	
				if (decpkt(rcvbuf, n, rcv))
					continue;
				else if ( reqid == rcv->pdu->reqid ) {
					freebytes(snmp);
					snmpdbg(DEBUG, MINUS, "snmpTalk()", "pkt readed id = %d", reqid);
					free(rcvbuf);
					close(fd);
					alarm(0);
					return (0);
				}
		} while(!alarmflag);

		if (alarmflag)
				retries--;
		
	} while (retries);
	
	close(fd);
	free(rcvbuf);
	freebytes(snmp);
	alarm(0);
	return(-1);
}

/* ************************* */
/* Decodification Functions */
/* ************************* */

int
decpkt(uchar *buff, int bufflen, Packet *pkt)
{
	Bytes *bbuff, *bcomm;
	Elem emess, *ever, *ecomm, pdu, *epdu;
	Elist *elmess, *el, *elpdu;
	char *comm;
	int lcver;

	snmpdbg(DEBUG, PLUS, "decpkt()", "Decodifing the Packet.");
	
	bbuff = makebytes(buff, bufflen);

	if (decode(bbuff->data, bbuff->len, &emess) != ASN_OK) {
		snmpdbg(DEBUG, MINUS, "decpkt()", "Error in ASN.1 decode.");
		freebytes(bbuff);
		return(-1);
	}

	freebytes(bbuff);

	if( !is_seq(&emess, &elmess)  ) {
		snmpdbg(DEBUG, MINUS, "decpkt()", "Cannot recognize message.");
		freevalfields(&emess.val);
		return(-2);
	}

	if ( elistlen(elmess) != 3 ) {
		snmpdbg(DEBUG, MINUS, "decpkt()", "Cannot recognize message: n = %d", elistlen(elmess));
		freevalfields(&emess.val);
		return(-3);
	}

	/* Get the version */
	ever = &elmess->hd;
	if ( !is_int(ever, &lcver) ) {
		snmpdbg(DEBUG, MINUS, "decpkt()", "Error decoding version field.");
		freevalfields(&emess.val);
		return(-4);
	}

	pkt->version = lcver;

	el = elmess->tl;

	/* Get the community */
	ecomm = &el->hd;
	if ( !is_octetstring(ecomm, &bcomm) ) {
		snmpdbg(DEBUG, MINUS, "decpkt()", "Error decoding community field.");
		freevalfields(&emess.val);
		return(-5);
	}

	/* Save the community */
	comm = (char *) malloc((bcomm->len)+1);
	memcpy(comm, (char*)bcomm->data, bcomm->len);
	comm[bcomm->len]='\0';

	/* Assign to the packet, the memory will be freeded en freepkt() */
	pkt->comm = comm;

	snmpdbg(DEBUG, EQUAL, "decpkt()", "Version = %d Community = %s", pkt->version, pkt->comm);

	/* Move the pointer */
	el = el->tl;
	epdu = &el->hd;

	if ( pdu2asn(epdu, &pdu) ) {
		snmpdbg(DEBUG, MINUS, "decpkt()", "Error decoding pdu.");
		freevalfields(&emess.val);
		return(-5);
	}

	if ( !is_seq(&pdu, &elpdu) ) {
		snmpdbg(DEBUG, MINUS, "decpkt()", "Error decoding pdu.");
		freevalfields(&emess.val);
		return(-7);
	}

	if ( decpdu(elpdu, pkt) != ASN_OK ) {
		snmpdbg(DEBUG, MINUS, "decpkt()", "Error decoding pdu field.");
		freevalfields(&emess.val);
		return(-8);
	}

	freevalfields(&pdu.val);
	freevalfields(&emess.val);

	snmpdbg(DEBUG, MINUS, "decpkt()", "Packet decodification Success.");

	return(0);
}


int
decpdu(Elist *el, Packet *pkt)
{
	Elem *elem, last;
	Elist *pel = nil, *elvar = nil, *elbind = nil;
	Pdu *pdu;
	int err, tabsize = 0, i;
	Tuple *storetpl = nil, *tmptpl;
	int lcreqid, lcerrstat, lcerridx;
	char *tmpchar1, *tmpchar2, *tmpchar3;

	pdu = mkpdu();

	if (pdu == nil) {
		snmpdbg(DEBUG, MINUS, "dec_pdu()", "Not PDU created.");
		return(-2);
	}

	snmpdbg(DEBUG, PLUS, "dec_pdu()", "Decodifing a PDU.");

	elem = &el->hd;
	err = ASN_OK;

	/* Decode the reqid */
	if( is_int(elem, &lcreqid) ) {
		pel = el->tl;
		elem = &pel->hd;
	} else {
		snmpdbg(DEBUG, MINUS, "dec_pdu()", "Reqid not found.");
		err = ASN_EINVAL;
	}

	pdu->reqid = lcreqid;

	/* Decode the errstat */
	if ((err == ASN_OK) && is_int(elem, &lcerrstat)) {
		pel = pel->tl;
		elem = &pel->hd;
	} else {
		snmpdbg(DEBUG, MINUS, "dec_pdu()", "Errstat not found.");
		err = ASN_EINVAL;
	}

	pdu->errstat = lcerrstat;

	/* Decode the errindex */
	if ((err == ASN_OK) && is_int(elem, &lcerridx)) {
		pel = pel->tl;
		elem = &pel->hd;
	} else {
		snmpdbg(DEBUG, MINUS, "dec_pdu()", "Erridx not found.");
		err = ASN_EINVAL;
	}

	pdu->erridx = lcerridx;

	/* As debbuging info, print the values */
	if (err == ASN_OK) {
		snmpdbg(DEBUG, EQUAL, "dec_pdu()", "reqid = %d errstat = %d erridx = %d",
			pdu->reqid, pdu->errstat, pdu->erridx);
	}

	/* Decode the Varbind and get the number of elements */
	if ( (is_seq(elem, &elvar)) && (err == ASN_OK) ) {
		elem = &elvar->hd;
		tabsize = elistlen(elvar);
		snmpdbg(DEBUG, EQUAL, "dec_pdu()", "vlen = %d", tabsize);
	} else {
		snmpdbg(DEBUG, MINUS, "dec_pdu()", "Error decoding var list.");
		err = ASN_EINVAL;
	}

	/* Decode each oid */
	if (err == ASN_OK) {
		for(i = 0; i < tabsize; i++) {
			if ( is_seq(elem, &elbind) ) {
				pel = elbind;
				last = pel->hd;
				pel = pel->tl;
				if ( last.val.tag == VObjId ) {
					/* Se añade la PDU */

					tmpchar1 = asn2char(last);
					tmpchar2 = asn2char(pel->hd);
					
					tmpchar3 = smprint("%d", pel->hd.tag.class);

					if ((tmpchar1 != nil) && (tmpchar2 != nil)) {

						if (i == 0) {
							/* First oid */
							/* Must be one oid */
							storetpl = mktuple(tmpchar1, tmpchar2, tmpchar3, 2);
						} else {
							/* The next oids */
							tmptpl = mktuple(tmpchar1, tmpchar2, tmpchar3, 2);
							linktuples(storetpl, tmptpl);
						}

						free(tmpchar1);
						free(tmpchar2);
						free(tmpchar3);
					} else {
						snmpdbg(DEBUG, MINUS, "dec_pdu()", "Any value is nil.");
						return ASN_EINVAL;
					}
				} else {
					snmpdbg(DEBUG, MINUS, "dec_pdu()", "Error decoding var list.");
					return ASN_EINVAL;
				}
			} else {
				snmpdbg(DEBUG, MINUS, "dec_pdu()", "Error decoding bind list.");
				err = ASN_EINVAL;
			}
			/* Prepare to the next */
			elvar = elvar->tl;
			elem = &elvar->hd;
		}
		pdu->tuple = storetpl;
	}

	if (err == ASN_OK) {
		pkt->pdu = pdu;

		snmpdbg(DEBUG, MINUS, "dec_pdu()", "PDU Decodification Success.");
	}
	
	return err;
}