#include <u.h>
#include <libc.h>
#include <bio.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include "../lib/asn1.h"
#include "../lib/snmp_asn1.h"
#include "../lib/snmp_debug.h"
/* *****************
  * Function file for snmp from/to asn1
  * ***************** */

/* *****************
  * Create Elem'ents
  * ***************** */
Elem
mkelem(int type, Elist *el)
{
	Elem e;

	e.tag.class = type;
	e.tag.num = SEQUENCE;
	e.val.tag = VSeq;
	e.val.u.seqval = el;
	return e;
}

/* *****************
  * Convert a string to a Int
  * ***************** */
Ints *
snmp_mkoid(char *str)
{
	char *f[32];
	int i, n;
	Ints *oid;

	n = getfields(str, f, 32, 0, ".");
	if ( n < 0 ) {
		oid = nil;
	} else {
		oid = (Ints*) emalloc9p(sizeof(Ints) * n);
		if (oid != nil) {
			oid->len = n;
			for(i=0; i<n; i++)
				oid->data[i] = atoi(f[i]);
		}
	}
	return oid;
}


/* *****************
  * Convert an asn string Elem to char
  * ***************** */
char *
asn2char(Elem e)
{
	Value v;
	int i, n, len;
	char *str = nil, *aux, *ptr;
	mpint *m;
	
	v = e.val;

	/* chek if i have a valid type before trying to decode it */
	if ( e.tag.class > 30 ) {
		switch(e.tag.class) {
			case 	IpAddress:

				str = (char*) emalloc9p(v.u.octetsval->len);
				memcpy(str, v.u.octetsval->data, v.u.octetsval->len);
				str = (char*)erealloc9p(str,v.u.octetsval->len+1);
				str[v.u.octetsval->len] = '\0';

				break;
			case Counter:
			case Gauge:  
			case TimeTicks:
			case Opaque:
			case NsapAddress:
			case Counter64:
			case UIInteger32:
				m = betomp(v.u.octetsval->data, v.u.octetsval->len, nil);
				aux = mptoa(m,10, nil, 0);
				str = smprint("%s",aux);
				mpfree(m);
				free(aux);
				break;
			case NoSuchObject:
				str = smprint("No Such Object");
				break;
			case NoSuchInstance:
				str = smprint("No Such Instance");
				break;
			case EndOfMIB:
				str = smprint("End of MIB");
				break;
			default:
				str = smprint("unknown class type %x and val tag %x", e.tag.class, e.val.tag);
				break;
		}
	
	} else {

		/* TODO: complete all known types */
		if ( e.tag.class == Universal ) {
			switch(v.tag) {
				case VBigInt:
					str = smprint("BINGINT");
					break;
				case VBitString:
					str = smprint("BITSTRING");
					break;
				case VBool:
					str = smprint("BOOLEAN");
					break;
				case VEOC:
					str = smprint("EOC");
					break;
				case VInt:
					str = smprint("%d", v.u.intval);
					break;
				case VNull:
					str = smprint("NULL");
					break;
				case VReal:
					str = smprint("REAL");
					break;
				case VObjId:
					len = 0;
					str = (char*)emalloc9p(1);
					for(i = 0; i<v.u.objidval->len; i++) {
						aux = smprint("%d%s", v.u.objidval->data[i],(i != (v.u.objidval->len-1)) ? ".":"");
						n = strlen(aux);
						len += n;
						str = (char*)erealloc9p(str,len);
						ptr = str+(len-n);
						memcpy(ptr,aux,n);
						free(aux);
					}
					str = (char*)erealloc9p(str,len+1);
					str[len] = '\0';
					break; 
				case VOctets:
					str = (char*) emalloc9p(v.u.octetsval->len);
					memcpy(str, v.u.octetsval->data, v.u.octetsval->len);
					str = (char*)erealloc9p(str,v.u.octetsval->len+1);
					str[v.u.octetsval->len] = '\0';
					break;
				case VOther:
					str = smprint("OTHER");
					break;
				case VSeq:
					str = smprint("SEQUENCE");
					break;
				case VSet:
					str = smprint("SET");
					break;
				case VString:
					str = smprint("%s", v.u.stringval);
					break;
				default:
					str = smprint(str,"Class %d - Type %d is not decoded", e.tag.class, v.tag);
					break;
			}		
		} else {
			str = smprint("asn2char(): Class is not UNIVERSAL %x", e.tag.class);
		}
	}
	return str;
}

/* ----
  * getBulkRequest PDU ASN1
  *	Request-ID INTEGER
  *	Non-Repeaters INTEGER
  *	Max-Repetition INTEGER
  *	Var-Bind-List =SEQUENCE 		
  *
  * OR
  *
  * getNextRequest PDU ASN1				
  *		Request-ID INTEGER			
  *		Error-Status INTEGER			
  *		Error-Index INTEGER			
  *		Var-Bind-List =SEQUENCE 		
  * -----*/
Elist *
mkrequest(int type, Elem reqid, Elem f1, Elem f2, Elem el)
{
	Elist *list1, *list2, *list3, *list4, *list5;
	Elem e3;

	list1 = mkel(el, nil);
	list2 = mkel(f2, list1);
	list3 = mkel(f1, list2);
	list4 = mkel(reqid, list3);
	e3 = mkelem((int) type, list4);
	list5 = mkel(e3, nil);

	return list5;
}


/* 
  * the asn.1 thing of libsec stores unknown objects as OCTECT_STRING
  * see ber_decode() function in asn1.h, we will take those raw bytes and
  * will construct a sequence adding the sequence ID and lenght of the
  * bytes encoded  at the begining of the bytes in eepdu, then
  * we decode as a normal sequence
  */

int
pdu2asn(Elem *epdu, Elem *e)
{
	int err;
	Bytes *ocpdu;
 
	if ( encode(*epdu,&ocpdu) == ASN_OK ) {
		*ocpdu->data = 0x30; /* pdu is a sequence, so we make it that */
	} else {
		snmpdbg(DEBUG, MINUS, "pdu2asn()", "Error encoding Octet-PDU.");
		freebytes(ocpdu);
		return(-1);
	}
	 
	/* decode what we encoded, to obtain the pdu */
	err = decode(ocpdu->data, ocpdu->len, e);
	freebytes(ocpdu);
	
	if ( err != ASN_OK ) {
		snmpdbg(DEBUG, MINUS, "pdu2asn()", "Error decoding PDU.");
		return(-2);
	}

	return(0);
}


Elem
oid2asn(char *s)
{
	Elem e;
	Ints *o;
	char *f[MAX_ELEM];
	int i, n;
	char *aux;

	aux = strdup(s);
	n = getfields(aux, f, MAX_ELEM, 0, ".");

	if ( n < 0 || atoi(f[0]) < 1 ) {
		free(aux);
		return e;
	}

	o = (Ints*) emalloc9p(sizeof(Ints)*n);
	o->len = n;
	for(i=0;i<n;i++)
		o->data[i] = atoi(f[i]);
	
	e = mkoid(o);

	snmpdbg(DEBUG, EQUAL, "oid2asn()", "oid = %s o->len = %d", s ,o->len);

	free(aux);
	freeints(o);
	return e;
}


Elem
char2asn(char *data, char *type)
{
	/* for normal requests */
	if ((data == nil) || (type == nil))
		return Null();

	/* Intenger */
	if (!strcmp("i", type)) {
		snmpdbg(DEBUG, GOBACK, "char2asn()", "Integer Element = %s", data);
		return mkint(atoi(data));
	}

	/* Octect-String */
	if (!strcmp("o", type)) {
		snmpdbg(DEBUG, GOBACK, "char2asn()", "Octect-String Element");
		return mkoctet((uchar*)data, strlen(data));	
	}

	/* String */
	if (!strcmp("s", type)) {
		snmpdbg(DEBUG, GOBACK, "char2asn()", "String Element = %s", data);
		return mkstring((char*)data);	
	}

	/* Bits */
	if (!strcmp("b", type)) {
		snmpdbg(DEBUG, GOBACK, "char2asn()", "Bits Element");
		return mkbits((uchar*)data, strlen(data));
	}

	/* UTC */
	/* Not implemented:

		Elem mkbigint(mpint *p);
		Elem mkutc(long t);
		Elem mkoid(Ints *oid);
		Elem mkseq(Elist *el);
		Elem mkset(Elist *el);
		Elem mkDN(char *dn);
	*/

	snmpdbg(DEBUG, GOBACK, "char2asn()", "Type is not supported");
	return mkint(atoi(data));
}
